#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/nodes/SoCoordinate4.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoTexture2.h>

#include "marchCube.h"


////////////////////////////////////////////////////////////
//
// Kevin Beason 4/18/2000
// Based on code by David C Banks
//
// Creates a water and ship isosurface.
//
// My depiction of the grid:
//
//       +--------+
//      /        /|
//     /        / |
//    +--------+  |
//    |        |  |
//  k |        |  +
//    |        | / 
//    |        |/ j
//    +--------+
//        i
//

//#define USEREDUCED
#define NOWINDOW

const double xdomain=2.0;
const double ydomain=0.5;
const double zdomain=0.5;
const double xblob=1.0-0.5;
const double yblob=0.25;
const double zblob=0.25;
const int localnx=128;
const int localny=96;
const int localnz=96;
typedef double vec3f[3];
vec3f bxmin, bxmax;
int bimax, bjmax, bkmax;

#ifndef USEREDUCED
const int imax=localnx;
const int jmax=localny;
const int kmax=localnz;
const int ijkdim=imax*jmax*kmax;
const int NGROW=3;
const int bijkdim=129*65*65;
char *waterInFileName  = "/home/beason/sussman/check.good";
char *waterOutFileName = "/home/beason/sussman/check.qlevel2";
char *boatInFileName  = "/home/beason/sussman/eblevel.dat";
char *boatOutFileName = "/home/beason/sussman/eblevel.dat.bin";
#endif

#ifdef USEREDUCED
const int imax=localnx/2;
const int jmax=localny/2;
const int kmax=localnz/2;
const int ijkdim=imax*jmax*kmax;
const int NGROW=0;
const int bijkdim=129*65*65;
char *waterInFileName  = "/home/beason/sussman/check.2";
char *waterOutFileName = "/home/beason/sussman/check.2.qlevel2";
char *boatInFileName  = "/home/beason/sussman/eblevel.dat.2";
char *boatOutFileName = "/home/beason/sussman/eblevel.dat.2.bin";
#endif

int index(int a, int i, int j, int k){
  if (a==1){
    if (i<0 || j<0 || k<0 || i>=imax || j>=jmax || k>=kmax)
      cerr << "index1() arguments out of bounds! "
           << "i=" << i << " j=" << j << " k=" << k << "\n";
    return (jmax*kmax*i + kmax*j + k);
  }
  if (i<0 || j<0 || k<0 || i>=bimax || j>=bjmax || k>=bkmax)
    cerr << "index1() arguments out of bounds! "
         << "i=" << i << " j=" << j << " k=" << k << "\n";
  return(bjmax*bkmax*i + bkmax*j + k);
}

inline double xco(int a, int i){
  if (a==1) return((double)i * (xdomain / (double) imax));

  // map 0..bimax to bxmin[0]..bxmax[0]
  return((bxmax[0]-bxmin[0])/(double)bimax * (double) i + bxmin[0] +xblob);
}

inline double yco(int a, int j){
  if (a==1) return((double)j * (ydomain / (double) jmax));
  return((bxmax[1]-bxmin[1])/(double)bjmax * (double) j + bxmin[1] +yblob);
}

inline double zco(int a, int k){
  if (a==1) return((double)k * (zdomain / (double) kmax));
  return((bxmax[2]-bxmin[2])/(double)bkmax * (double) k + bxmin[2] +zblob);
}

double cornerLvl[8];
void fillCorners(int a, int i, int j, int k, double levels[]){
  cornerLvl[0]=levels[index(a,i,j,k)];
  cornerLvl[1]=levels[index(a,i+1,j,k)];
  cornerLvl[2]=levels[index(a,i+1,j+1,k)];
  cornerLvl[3]=levels[index(a,i,j+1,k)];
  
  cornerLvl[4]=levels[index(a,i,j,k+1)];
  cornerLvl[5]=levels[index(a,i+1,j,k+1)];
  cornerLvl[6]=levels[index(a,i+1,j+1,k+1)];
  cornerLvl[7]=levels[index(a,i,j+1,k+1)];
  return;
}

void setVerts(int a, int i, int j, int k, SoCoordinate4* g){
  // make sure you called fillCorners()!
  g->point.set1Value(0,xco(a,i),yco(a,j),zco(a,k),      cornerLvl[0]);
  g->point.set1Value(1,xco(a,i+1),yco(a,j),zco(a,k),    cornerLvl[1]);
  g->point.set1Value(2,xco(a,i+1),yco(a,j+1),zco(a,k),  cornerLvl[2]);
  g->point.set1Value(3,xco(a,i),yco(a,j+1),zco(a,k),    cornerLvl[3]);

  g->point.set1Value(4,xco(a,i),yco(a,j),zco(a,k+1),    cornerLvl[4]);
  g->point.set1Value(5,xco(a,i+1),yco(a,j),zco(a,k+1),  cornerLvl[5]);
  g->point.set1Value(6,xco(a,i+1),yco(a,j+1),zco(a,k+1),cornerLvl[6]);
  g->point.set1Value(7,xco(a,i),yco(a,j+1),zco(a,k+1),  cornerLvl[7]);
  return;
}


SoSeparator *
makeScene ()
  {
  SoSeparator *root = new SoSeparator();
  return root;
  }


void
main(int, char **argv)
  {

  // Inventer initialization code
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);
  
  // Create a scene graph
  SoSeparator *root = makeScene();


  //======================================================================
  // Eric Hernandez's Code
  //
  // Sets up camera to front of boat
  //======================================================================
  
  SoTransform* SceneTransformA = new SoTransform;
  SceneTransformA->translation.setValue(0.0, 0.0, 100.0);

  SoTransform* SceneTransform = new SoTransform;
  SceneTransform->rotation.setValue(SbVec3f(0.0, 0.0, 1.0), 1.80);

  SoTransform* SceneTransform2 = new SoTransform;
  SceneTransform2->rotation.setValue(SbVec3f(-1.0, 0.0, 0.0), 1.40);

  SoTransform* SceneTransform3 = new SoTransform;
  SceneTransform3->translation.setValue(-1.0, -0.25, -0.25);

  root->addChild(SceneTransform3);
  root->addChild(SceneTransform2);
  root->addChild(SceneTransform);
  root->addChild(SceneTransformA);

  //======================================================================

  

  // Give Inventor some help in rendering my polygons
  SoShapeHints *twoSided = new SoShapeHints();
  twoSided->shapeType = SoShapeHints::UNKNOWN_SHAPE_TYPE;
  twoSided->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
  root->addChild(twoSided);


  //======================================================================
  // Read data from water file
  //======================================================================

  double qvfrac, qlevel, qvel1, qvel2, qvel3, wsoln;
  int i, j, k;
  double qlevels[ijkdim];
  double qlevelmax=0, qlevelmin=0;
  const int readFile=1;

  // open input file
  cout << "Opening water input file: " << waterInFileName << endl;
  ifstream waterInFile;
  waterInFile.open(waterInFileName);

  // open output file
  cout << "Opening water output file: " << waterOutFileName << endl;
  FILE *waterOutFile;
  waterOutFile=fopen(waterOutFileName, "wb");

  // read in data
  waterInFile >> qvfrac >> qlevel; // trash first 2 values
  cout << "Reading i level (" << -NGROW << "=>" << imax+NGROW << ")... ";
  if (readFile){
    for (i=-NGROW;i<imax+NGROW;i++){
      cout << i << " " << flush;
      for (j=-NGROW;j<jmax+NGROW;j++){
        for (k=-NGROW;k<kmax+NGROW;k++){
	  waterInFile >> qvfrac >> qlevel >> qvel1 >> qvel2 >> qvel3 >> wsoln;
          if (i<0 || i>=imax || j<0 || j>=jmax || k<0 || k>=kmax) continue;
  	  qlevels[index(1,i,j,k)] = qlevel;
 	  fwrite(&qlevel, sizeof(qlevel), 1, waterOutFile);
          if (qlevel<qlevelmin) qlevelmin=qlevel;
	  if (qlevel>qlevelmax) qlevelmax=qlevel;
        }
      }
    }
  }
  cout << endl;
  cout << "qlevelmin = " << qlevelmin << endl
       << "qlevelmax = " << qlevelmax << endl;

  // close files
  cout << "Closing files: " << endl;
  cout << waterInFileName << endl;
  waterInFile.close();  
  cout << waterOutFileName << endl;
  fclose(waterOutFile);

  //======================================================================
  // Read data from boat file
  //======================================================================

  // read in data
  double boatlevel, trash;
  double boatlevels[bijkdim];

  // open input file
  cout << "Reading from data file: " << boatInFileName << endl;
  ifstream boatInFile;
  boatInFile.open(boatInFileName);
  
  // open output file
  cout << "Writing to data file: " << boatOutFileName << endl;
  FILE *boatOutFile;
  boatOutFile=fopen(boatOutFileName, "wb");
  
  boatInFile >> bimax >> bjmax >> bkmax;
  cout << "bimax= " << bimax
       << " bjmax= " << bjmax
       << " bkmax= " << bkmax << endl;
  boatInFile >> trash >> trash >> trash;
  boatInFile >> bxmin[0] >> bxmin[1] >> bxmin[2];
  boatInFile >> bxmax[0] >> bxmax[1] >> bxmax[2];

  // loop through input data and write out, in binary
  cout << "Reading i = (0.." << bimax << ") ";
  for (i=0;i<bimax;i++){
    cout << i << " " << flush;
    for (j=0;j<bjmax;j++)
      for (k=0;k<bkmax;k++){
	boatInFile >> boatlevel;
	boatlevels[index(2,i,j,k)]=boatlevel;
	fwrite(&boatlevel, sizeof(boatlevel), 1, boatOutFile);
      }
  }

  cout << "\nClosing files: " << endl;
  cout << boatInFileName << endl;
  boatInFile.close();
  cout << boatOutFileName << endl;
  fclose(boatOutFile);


  //======================================================================
  // Create Isosurface for water
  //======================================================================

  const int maxNumGrids=jmax*kmax*5*2; // 46080
  const int vertsPerGrid=15;
  const int facesPerGrid=20;
  int currGridNum=0;
  int gi, faceValue;
  SoCoordinate3    *isoVerts       = new SoCoordinate3();
  SoCoordinate3    *globalVerts    = new SoCoordinate3();
  SoCoordinate3    *globalVerts2   = new SoCoordinate3();
  SoIndexedFaceSet *isoFaceSet     = new SoIndexedFaceSet();
  SoIndexedFaceSet *globalFaceSet  = new SoIndexedFaceSet();
  SoIndexedFaceSet *globalFaceSet2 = new SoIndexedFaceSet();
  SoCoordinate4    *gridVerts      = new SoCoordinate4();
  
  //======================================================================
  // Eric Hernandez's code
  //
  // Create material for water surface (transparent)
  //======================================================================

  SoMaterial *myMaterial = new SoMaterial();
  myMaterial->transparency.setValue(0.5);
  myMaterial->diffuseColor.setValue(0.3, 0.3, 1.0);
  //myMaterial->ambientColor.setValue(0.92, 1.0, 0.14);
  //myMaterial->specularColor.setValue(0.95, 0.03, 1.0);
  //myMaterial->emissiveColor.setValue(0.0, 0.0, 0.0);
  //myMaterial->shininess.setValue(.5);
  root->addChild( myMaterial );

  //======================================================================
  // loop through data and run marching cubes
  cout << "Running marching-cubes for water surface, i= (0=>"
       << imax << ") " << flush;
  for (i=0;i<imax-1;i++){
    cout << i << " " << flush;
    for (j=0;j<jmax-1;j++)
      for(k=0;k<kmax-1;k++){
	fillCorners(1,i,j,k,qlevels);
	if(cornerLvl[0]*cornerLvl[1]<0.0 ||
	   cornerLvl[0]*cornerLvl[2]<0.0 ||
	   cornerLvl[0]*cornerLvl[3]<0.0 ||
	   cornerLvl[0]*cornerLvl[4]<0.0 ||
	   cornerLvl[0]*cornerLvl[5]<0.0 ||
	   cornerLvl[0]*cornerLvl[6]<0.0 ||
	   cornerLvl[0]*cornerLvl[7]<0.0){   // isosurface at isovalue 0.0
                                             // detected

	  // add surface for this cube
          setVerts(1,i,j,k,gridVerts);

	  // create vertices
	  isoVerts->point.setNum(vertsPerGrid);
	  for (gi=0;gi<vertsPerGrid;gi++)
	    isoVerts->point.set1Value(gi,0.0,0.0,0.0);

	  // create faces
	  isoFaceSet->coordIndex.setNum(facesPerGrid);
	  for (gi=0;gi<facesPerGrid;gi++)
	    isoFaceSet->coordIndex.set1Value(gi,0);

	  // set vertices + faces using marching cubes
	  polygonize(gridVerts->point.getValues(0), 0.0, 
		     isoVerts->point.startEditing(),
		     isoFaceSet, root);
	  isoVerts->point.finishEditing();

	  // Add stuff to scenegraph
	  //root->addChild(isoVerts);
	  globalVerts->point.setNum((currGridNum+1)*vertsPerGrid);
	  for (gi=0;gi<vertsPerGrid;gi++)
	    globalVerts->point.set1Value(currGridNum*vertsPerGrid+gi,
					 isoVerts->point[gi]);
	  //root->addChild(isoFaceSet);
 	  globalFaceSet->coordIndex.setNum((currGridNum+1)*facesPerGrid);
	  for (gi=facesPerGrid-1;gi>=0;gi--)
	    if (isoFaceSet->coordIndex[gi] != 0) break;
	  int lastValidIndex=gi;
	  for (gi=0;gi<=lastValidIndex;gi++){
	    faceValue=isoFaceSet->coordIndex[gi];
	    if (faceValue!=-1) faceValue+=currGridNum*vertsPerGrid;
	    globalFaceSet->coordIndex.set1Value(currGridNum*facesPerGrid+gi,
						faceValue);
	  }
	  for (gi=lastValidIndex+1;gi<facesPerGrid;gi++)
	  globalFaceSet->coordIndex.set1Value(currGridNum*facesPerGrid+gi,-1);
				
	  // detect if there are too many faces
	  currGridNum++;
	  if (currGridNum>=maxNumGrids){
	    cerr << "Maximum number of grids reached!\n" << flush;
	    goto endOfWaterCubesLoop;
	  }
	}
      }
  }
  endOfWaterCubesLoop:
  cout << endl;

  root->addChild(globalVerts);
  root->addChild(globalFaceSet);

  //======================================================================
  // Create Isosurface for boat
  //======================================================================

  currGridNum=0;
  
  // create material for surface
  SoMaterial *myBoatMaterial = new SoMaterial();
  myBoatMaterial->diffuseColor.setValue(1.0, 1.0, 1.0);
  root->addChild( myBoatMaterial );

  // loop through data and run marching cubes
  cout << "Running marching-cubes for boat surface, i= (0=>"
       << bimax << ") " << flush;
  for (i=0;i<bimax-1;i++){
    cout << i << " " << flush;
    for (j=0;j<bjmax-1;j++)
      for(k=0;k<bkmax-1;k++){
	fillCorners(2,i,j,k,boatlevels);
	if(cornerLvl[0]*cornerLvl[1]<0.0 ||
	   cornerLvl[0]*cornerLvl[2]<0.0 ||
	   cornerLvl[0]*cornerLvl[3]<0.0 ||
	   cornerLvl[0]*cornerLvl[4]<0.0 ||
	   cornerLvl[0]*cornerLvl[5]<0.0 ||
	   cornerLvl[0]*cornerLvl[6]<0.0 ||
	   cornerLvl[0]*cornerLvl[7]<0.0){   // isosurface at isovalue 0.0
                                             // detected
	       
	  // add surface for this cube
          setVerts(2,i,j,k,gridVerts);

	  // create vertices
	  isoVerts->point.setNum(vertsPerGrid);
	  for (gi=0;gi<vertsPerGrid;gi++)
	    isoVerts->point.set1Value(gi,0.0,0.0,0.0);

	  // create faces
	  isoFaceSet->coordIndex.setNum(facesPerGrid);
	  for (gi=0;gi<facesPerGrid;gi++)
	    isoFaceSet->coordIndex.set1Value(gi,0);

	  // set vertices + faces using marching cubes
	  polygonize(gridVerts->point.getValues(0), 0.0, 
		     isoVerts->point.startEditing(),
		     isoFaceSet, root);
	  isoVerts->point.finishEditing();

	  // Add stuff to scenegraph
	  //root->addChild(isoVerts);
	  globalVerts2->point.setNum((currGridNum+1)*vertsPerGrid);
	  for (gi=0;gi<vertsPerGrid;gi++)
	    globalVerts2->point.set1Value(currGridNum*vertsPerGrid+gi,
					 isoVerts->point[gi]);
	  //root->addChild(isoFaceSet);
 	  globalFaceSet2->coordIndex.setNum((currGridNum+1)*facesPerGrid);
	  for (gi=facesPerGrid-1;gi>=0;gi--)
	    if (isoFaceSet->coordIndex[gi] != 0) break;
	  int lastValidIndex=gi;
	  for (gi=0;gi<=lastValidIndex;gi++){
	    faceValue=isoFaceSet->coordIndex[gi];
	    if (faceValue!=-1) faceValue+=currGridNum*vertsPerGrid;
	    globalFaceSet2->coordIndex.set1Value(currGridNum*facesPerGrid+gi,
						faceValue);
	  }
	  for (gi=lastValidIndex+1;gi<facesPerGrid;gi++)
	  globalFaceSet2->coordIndex.set1Value(currGridNum*facesPerGrid+gi,-1);

	  // detect if there are too many faces
	  currGridNum++;
	  if (currGridNum>=maxNumGrids){
	    cerr << "Maximum number of grids reached!\n" << flush;
	    goto endOfBoatCubesLoop;
	  }
	}
      }
  }
  endOfBoatCubesLoop:
  cout << endl;

  root->addChild(globalVerts2);
  root->addChild(globalFaceSet2);

  //======================================================================
  // Eric Hernandez's Code
  //
  // Add's big water surface
  //======================================================================
  
  SoSeparator* WaterRoot = new SoSeparator();

  SoTexture2 *WaterTexture = new SoTexture2();
  WaterTexture->filename.setValue(SbString("water.rgb"));
  WaterTexture->model = SoTexture2::MODULATE;
  WaterTexture->wrapS = SoTexture2::CLAMP;
  WaterTexture->wrapT = SoTexture2::CLAMP;
  WaterRoot->addChild(WaterTexture);
 
  SoMaterial* WaterMaterial = new SoMaterial();
  //WaterMaterial->ambientColor.setValue(SbColor(.0,.0,.0));
  WaterMaterial->diffuseColor.setValue(SbColor(0.2, 0.2, 1.0));
  WaterMaterial->transparency.setValue(0.5);
  WaterRoot->addChild(WaterMaterial);  

  SoTransform* WaterTransform = new SoTransform();
  WaterTransform->scaleFactor.setValue(3.0, 3.0, 0.001);
  WaterTransform->translation.setValue(1.0, 0.2, 0.235);
  WaterRoot->addChild(WaterTransform);

  SoCube* Water = new SoCube();
  WaterRoot->addChild(Water);

  root->addChild(WaterRoot);

  //======================================================================

  
  //======================================================================
  // Open a Viewer
  //======================================================================

#ifndef NOWINDOW
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->setTitle("Boat");
  myViewer->show();
#endif

  //==================================================================
  // Write to file
  //==================================================================

  SoWriteAction myWriteAction;
  myWriteAction.getOutput()->openFile("/home/beason/sussman/myiso/scene.iv");
  myWriteAction.getOutput()->setBinary(FALSE);
  myWriteAction.apply(root);
  myWriteAction.getOutput()->closeFile();

  //system("/home/beason/sussman/isosurf/addscene.pl");

  //======================================================================
  // Pass Control to Viewer
  //======================================================================

#ifndef NOWINDOW
  SoXt::show(myWindow);
  SoXt::mainLoop();
#endif

}


